﻿using System;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Windows.Forms;

namespace Conv2PSV
{
    class Program
    {
        static string GetString(Stream fs)
        {
            String str = "";
            byte by = 0xFF;
            while (true) 
            {
                by = (byte)fs.ReadByte();
                if (by == 0x00)
                    break;
                str += (char)by;
            }
            
            return str;
        }

        static void WriteMagic(Stream fs)
        {
            fs.WriteByte(0x00);
            WriteString(fs, "VSP");
        }

        static void WriteString(Stream fs, String str)
        {
            char[] CharArray = str.ToArray();
            foreach(char Chr in CharArray)
            {
                fs.WriteByte((byte)Chr);
            }
        }
        
        static String GetTitle(String FileName)
        {
            if (Path.GetExtension(FileName) == ".mcs")
            {
                FileStream fs = File.OpenRead(FileName);
                fs.Seek(0xA, SeekOrigin.Begin);
                String Name = GetString(fs);
                fs.Close();
                return Name;
            }
            else if(Path.GetExtension(FileName) == ".gme")
            {
                FileStream fs = File.OpenRead(FileName);
                fs.Seek(0xFCA, SeekOrigin.Begin);
                String Name = GetString(fs);
                fs.Close();
                return Name;
            }
            else
            {
                throw new FileNotFoundException();
            }
        }

   
        static byte[] SignatureGen(Stream fs)
        {
            byte[] HMAC_KEY = { 0x02, 0x62, 0x8C, 0x13, 0x94, 0x71, 0xF5, 0x0B, 0xC8, 0x94, 0x4E, 0x6B, 0xF0, 0xDC, 0xBC, 0x50, 0x30, 0x3F, 0x1F, 0x5B }; // Key for this specific seed,

            fs.Seek(0x00, SeekOrigin.Begin);
            Console.WriteLine("Generating HMAC signature...");
            HMACSHA1 hmac = new HMACSHA1(HMAC_KEY);
            byte[] Signature = hmac.ComputeHash(fs);
            Console.WriteLine(BitConverter.ToString(Signature).Replace("-"," "));
            return Signature;
        }

        static String GetPsvName(String SaveName)
        {
            String PsvName = "";
            char[] SaveArray = SaveName.ToArray();
            bool DashFound = false;
            foreach(char SaveChar in SaveArray)
            {
                if(!DashFound)
                {
                    if (SaveChar == '-')
                    {
                        DashFound = true;
                    }
                    PsvName += SaveChar;
                }
                else
                {
                    if (Char.IsDigit(SaveChar))
                    {
                        PsvName += SaveChar;
                    }
                    else
                    {
                        PsvName += String.Format("{0:X}", Convert.ToInt32(SaveChar));
                    }
                }

            }
            PsvName += ".PSV";
            return PsvName;
        }

        static void WriteSC(String FilePath, Stream PSV)
        {
            Console.WriteLine("Writing SC Image");
            if (Path.GetExtension(FilePath) == ".mcs")
            {
                FileStream MCS = File.OpenRead(FilePath);
                long SCLen = MCS.Length - 0x80;
                MCS.Seek(0x80, SeekOrigin.Begin);
                byte[] SCImage = new Byte[SCLen];
                MCS.Read(SCImage, 0x00, (int)SCLen);
                PSV.Write(SCImage, 0x00, (int)SCLen);
                MCS.Close();
            }
            else if (Path.GetExtension(FilePath) == ".gme")
            {
                FileStream GME = File.OpenRead(FilePath);
                GME.Seek(0xFC4, SeekOrigin.Begin);
                BinaryReader BGME = new BinaryReader(GME);
                int SCLen = BGME.ReadInt16();
                GME.Seek(0x2F40, SeekOrigin.Begin);
                byte[] SCImage = new Byte[SCLen];
                GME.Read(SCImage, 0x00, SCLen);
                PSV.Write(SCImage, 0x00, SCLen);
                GME.Close();
            }
            else
            {
                throw new FileNotFoundException();
            }
        }

        [STAThread]
        static void Main(string[] args)
        {
            if(args.Length == 0)
            {
                Console.WriteLine("Select Save");
                OpenFileDialog ofd = new OpenFileDialog();
                ofd.InitialDirectory = Directory.GetCurrentDirectory();
                ofd.Filter = "PlayStation One Save Files (*.mcs, *.gme)|*.mcs;*.gme|MemCardREX Single Save (*.mcs)|*.mcs|PlayStation DEXDRIVE (Datel) (*.gme)|*.gme";
                DialogResult res = ofd.ShowDialog();
                if (res == DialogResult.OK)
                {
                    args = new String[] { ofd.FileName };
                }
            }

            if (args.Length != 0)
            {
                String FilePath = args[0];
                String SaveTitle = GetTitle(FilePath);

                Console.Write("PsvName = ");
                String PsvName = Path.Combine(Path.GetDirectoryName(FilePath),"PS3","EXPORT","PSV");
                Directory.CreateDirectory(PsvName);
                PsvName = Path.Combine(PsvName, GetPsvName(SaveTitle));
                Console.WriteLine(Path.GetFileName(PsvName));

                FileStream PSV = File.Open(PsvName, FileMode.OpenOrCreate,FileAccess.ReadWrite);
                PSV.SetLength(0);
                BinaryWriter BW = new BinaryWriter(PSV);
                WriteMagic(PSV);
                BW.Write((UInt32)0);

                //Am lazy and dont want to implement sonys broken AES-CBC algorythm
                //Just do a seed i know the hmac key for
                //ps3 wont care.

                Byte[] StaticSeeed = { 0x42, 0x6C, 0x65, 0x73, 0x73, 0x65, 0x64, 0x20, 0x42, 0x65, 0x20, 0x7E, 0x20, 0x57, 0x69, 0x63, 0x63, 0x61, 0x6E, 0x73 };
                //"Blessed Be ~ Wiccans"(0x14) - Allways nice to add personality to your code where-ever possible. 

                PSV.Write(StaticSeeed, 0x00, 0x14);

                PSV.Write(new Byte[0x14], 0x00, 0x14);
                Console.WriteLine("Writing Flags... ");
                Byte[] Flags = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x14, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x84, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x03, 0x90, 0x00, 0x00};
                PSV.Write(Flags, 0x00, 0x34);

                int Padding = (0x20 - SaveTitle.Length);
                Console.WriteLine("Writing " + SaveTitle + " +" + Padding.ToString());
                WriteString(PSV, SaveTitle);
                BW.Write(new Byte[Padding], 0x00, Padding);
                WriteSC(FilePath, PSV);

                byte[] Signature = SignatureGen(PSV);
                Console.WriteLine("Writing Signature to file...");
                PSV.Seek(0x1C, SeekOrigin.Begin);
                PSV.Write(Signature, 0x00, 0x14);
                Console.WriteLine("Done!\n\nBlessed Be ~");
                PSV.Close();

            }


        }
    }
}
